import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Mapping, Optional, Self

from aiohttp import web

from .json import dump_json


class ConfigurationError(Exception):
    invalid_data: Mapping[str, Any]

    def __init__(self, invalid_data: Mapping[str, Any]) -> None:
        super().__init__(invalid_data)
        self.invalid_data = invalid_data


class InvalidAPIHandlerDefinition(Exception):
    pass


class UnknownImageReference(ValueError):
    """
    Represents an error for invalid/unknown image reference.
    The first argument of this exception should be the reference given by the user.
    """

    def __str__(self) -> str:
        return f"Unknown image reference: {self.args[0]}"


class ImageNotAvailable(ValueError):
    """
    Represents an error for unavailability of the image in agents.
    The first argument of this exception should be the reference given by the user.
    """

    def __str__(self) -> str:
        return f"Unavailable image in the agent: {self.args[0]}"


class UnknownImageRegistry(ValueError):
    """
    Represents an error for invalid/unknown image registry.
    The first argument of this exception should be the registry given by the user.
    """

    def __str__(self) -> str:
        return f"Unknown image registry: {self.args[0]}"


class InvalidImageName(ValueError):
    """
    Represents an invalid string for image name.
    """

    def __str__(self) -> str:
        return f"Invalid image name: {self.args[0]}"


class InvalidImageTag(ValueError):
    """
    Represents an invalid string for image tag and full image name.
    Image tag should be a string of form below

    ```
    <version-stringA>-<platform-tag-1A>-<platform-tag-2A>-....
    ```
    """

    def __init__(self, tag: str, full_name: str | None = None) -> None:
        super().__init__(tag, full_name)
        self._tag = tag
        self._full_name = full_name

    def __str__(self) -> str:
        if self._full_name is not None:
            return f"Invalid or duplicate image name tag: {self._tag}, full image name: {self._full_name}"
        else:
            return f"Invalid or duplicate image name tag: {self._tag}"


class ProjectMismatchWithCanonical(ValueError):
    """
    Represent the project value does not match the canonical value when parsing the string representing the image in ImageRef.
    """

    def __init__(self, project: str, canonical: str) -> None:
        super().__init__(project, canonical)
        self._project = project
        self._canonical = canonical

    def __str__(self) -> str:
        return f'Project "{self._project}" mismatch with the image canonical: {self._canonical}'


class AliasResolutionFailed(ValueError):
    """
    Represents an alias resolution failure.
    The first argument of this exception should be the alias given by the user.
    """

    def __str__(self) -> str:
        return f"Failed to resolve alias: {self.args[0]}"


class InvalidIpAddressValue(ValueError):
    """
    Represents an invalid value for ip_address.
    """


class VolumeMountFailed(RuntimeError):
    """
    Represents a mount process failure.
    """


class VolumeUnmountFailed(RuntimeError):
    """
    Represents a umount process failure.
    """


class ErrorDomain(enum.StrEnum):
    """
    An enum to represent the domain of the error.
    The domain is a string that represents the area where the error occurred.
    """

    BACKENDAI = "backendai"  # Whenever possible, use specific domain names instead of this one.
    API = "api"
    PLUGIN = "plugin"
    BGTASK = "bgtask"
    KERNEL = "kernel"
    USER = "user"
    SESSION = "session"
    GROUP = "group"
    DOMAIN = "domain"
    IMAGE = "image"
    IMAGE_ALIAS = "image-alias"
    TEMPLATE = "template"
    CONTAINER_REGISTRY = "container-registry"
    SCALING_GROUP = "scaling-group"
    INSTANCE = "instance"
    ENDPOINT = "endpoint"
    ENDPOINT_AUTO_SCALING = "endpoint-auto-scaling"
    ROUTE = "route"
    DOTFILE = "dotfile"
    VFOLDER = "vfolder"
    MODEL_SERVICE = "model-service"
    RESOURCE_PRESET = "resource-preset"
    STORAGE = "storage"
    AGENT = "agent"
    PERMISSION = "permission"
    METRIC = "metric"


class ErrorOperation(enum.StrEnum):
    """
    An enum to represent the operation where the error occurred.
    The operation is a string that represents the action that was being performed
    when the error occurred.
    """

    GENERIC = "generic"  # Whenever possible, use specific operation names instead of this one.
    CREATE = "create"
    ACCESS = "access"
    READ = "read"
    UPDATE = "update"
    START = "start"
    SOFT_DELETE = "soft-delete"
    HARD_DELETE = "purge"
    LIST = "list-query"
    AUTH = "auth"
    HOOK = "hook"
    REQUEST = "request"
    PARSING = "parsing"
    EXECUTE = "execute"
    SETUP = "setup"
    GRANT = "grant"


class ErrorDetail(enum.StrEnum):
    """
    An enum to represent the specific error that occurred during the operation.
    The error detail is a string that describes the specific error that occurred
    during the operation.
    """

    # Client Error
    BAD_REQUEST = "bad-request"
    NOT_FOUND = "not-found"
    # Conflict means the request conflicts with the current state of the server.
    CONFLICT = "conflict"
    # Already Exists means the resource already exists.
    ALREADY_EXISTS = "already-exists"
    # Invalid parameters means the received parameters are invalid.
    # This is different from BAD_REQUEST, which means the request is malformed.
    INVALID_PARAMETERS = "invalid-parameters"
    # Forbidden means the user is not allowed to access the resource.
    # This is different from UNAUTHORIZED, which means the user is not authenticated.
    FORBIDDEN = "forbidden"
    # Unauthorized means the user is not authenticated.
    # This means the user is not logged in or the token is invalid.
    UNAUTHORIZED = "unauthorized"

    # Server Error
    INTERNAL_ERROR = (
        "internal-error"  # Whenever possible, use specific error names instead of this one.
    )
    # UNAVAILABLE means the resource is not available.
    UNAVAILABLE = "unavailable"
    # UNREACHABLE means the resource is unreachable.
    UNREACHABLE = "unreachable"
    # TIMEOUT means the request timed out.
    TASK_TIMEOUT = "task-timeout"
    # DATA_EXPIRED means the data is expired.
    # This is different from NOT_FOUND, which means the resource does not exist.
    DATA_EXPIRED = "data-expired"
    CANCELED = "canceled"
    # Unexpected Error
    NOT_IMPLEMENTED = "not-implemented"
    DEPRECATED = "deprecated"
    # MISMATCH means the current state of the server does not match the expected state.
    # MISMATCH is used when the server is in a state that is not expected.
    MISMATCH = "mismatch"


@dataclass
class ErrorCode:
    """
    A dataclass to represent error codes for Backend.AI errors.

    All fields are lowercase.

    The top-level domain is set to backendai. Other domains represent the specific
    area where the error occurred, such as kernel, user, session, etc.
    Whenever possible, use detailed domain names instead of backendai of general purpose errors.

    The operation field represents the operation where the error occurred,
    such as create, delete, update, get, list, etc.

    The error_detail field describes the specific error that occurred during the operation.
    If it consists of two or more words, they should be connected with a hyphen (-).
    """

    domain: ErrorDomain
    operation: ErrorOperation
    error_detail: ErrorDetail

    @classmethod
    def default(cls) -> Self:
        """
        Returns the default error code for Backend.AI errors.
        This is used when the error code is not specified.
        The default error code is `backendai_generic_internal-error`.
        """
        return cls(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )

    def __str__(self) -> str:
        return f"{self.domain}_{self.operation}_{self.error_detail}"


class BackendAIError(web.HTTPError, ABC):
    """
    An RFC-7807 error class as a drop-in replacement of the original
    aiohttp.web.HTTPError subclasses.
    """

    error_type: str = "https://api.backend.ai/probs/general-error"
    error_title: str = "General Backend API Error."

    content_type: str
    extra_msg: Optional[str]

    body_dict: dict[str, Any]

    def __init__(self, extra_msg: str | None = None, extra_data: Optional[Any] = None, **kwargs):
        super().__init__(**kwargs)
        self.args = (self.status_code, self.reason, self.error_type)
        self.empty_body = False
        self.content_type = "application/problem+json"
        self.extra_msg = extra_msg
        self.extra_data = extra_data
        body = {
            "type": self.error_type,
            "title": self.error_title,
            "error_code": str(self.error_code()),
        }
        if extra_msg is not None:
            body["msg"] = extra_msg
        if extra_data is not None:
            body["data"] = extra_data
        self.body_dict = body
        self.body = dump_json(body)

    def __str__(self):
        lines = []
        if self.extra_msg:
            lines.append(f"{self.error_title} ({self.extra_msg})")
        else:
            lines.append(self.error_title)
        if self.extra_data:
            lines.append(" -> extra_data: " + repr(self.extra_data))
        return "\n".join(lines)

    def __repr__(self):
        lines = []
        if self.extra_msg:
            lines.append(
                f"<{type(self).__name__}({self.status}): {self.error_title} ({self.extra_msg})>"
            )
        else:
            lines.append(f"<{type(self).__name__}({self.status}): {self.error_title}>")
        if self.extra_data:
            lines.append(" -> extra_data: " + repr(self.extra_data))
        return "\n".join(lines)

    def __reduce__(self):
        return (
            type(self),
            (),  # empty the constructor args to make unpickler to use
            # only the exact current state in __dict__
            self.__dict__,
        )

    @classmethod
    @abstractmethod
    def error_code(cls) -> ErrorCode:
        """
        Returns the error code for this error.
        This is used in the API response to indicate the type of error.

        The error code is in the form {domain}_{operation}_{error}.
        For example, "kernel_create_invalid-image" or "kernel_create_timeout".
        """
        raise NotImplementedError("Subclasses must implement error_code() method.")


class MalformedRequestBody(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Malformed request body."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.PARSING,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class InvalidAPIParameters(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Invalid or Missing API Parameters."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.PARSING,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class ResourcePresetConflict(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/duplicate-resource"
    error_title = "Duplicate Resource Preset"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.RESOURCE_PRESET,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.CONFLICT,
        )


class MiddlewareParamParsingFailed(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Middleware parameter parsing failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.PARSING,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class ParameterNotParsedError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Parameter Not Parsed Error"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.PARSING,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class BgtaskNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/bgtask-not-found"
    error_title = "Background Task Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BGTASK,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class BgtaskFailedError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/bgtask-failed"
    error_title = "Background Task Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BGTASK,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class BgtaskCancelledError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/bgtask-cancelled"
    error_title = "Background Task Cancelled"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BGTASK,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.CANCELED,
        )


class UnreachableError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/unreachable"
    error_title = "Unreachable"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.UNREACHABLE,
        )


class PermissionDeniedError(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/permission-denied"
    error_title = "Permission Denied."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.AUTH,
            error_detail=ErrorDetail.FORBIDDEN,
        )



import asyncio
import hashlib
import logging
import socket
from dataclasses import dataclass
from typing import AsyncGenerator, Optional

import redis
from aiotools.server import process_index

from ai.backend.logging.utils import BraceStyleAdapter

from .. import redis_helper
from ..types import RedisConnectionInfo
from .queue import AbstractMessageQueue, MessageId, MQMessage

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_DEFAULT_AUTOCLAIM_IDLE_TIMEOUT = 300_000  # 5 minutes
_DEFAULT_AUTOCLAIM_INTERVAL = 60_000
_DEFAULT_AUTOCLAIM_COUNT = 64
_DEFAULT_QUEUE_MAX_LEN = 128


@dataclass
class RedisMQArgs:
    # Required arguments
    stream_key: str
    group_name: str
    node_id: str
    # Optional arguments
    autoclaim_idle_timeout: int = _DEFAULT_AUTOCLAIM_IDLE_TIMEOUT
    autoclaim_start_id: Optional[str] = None


class RedisQueue(AbstractMessageQueue):
    _conn: RedisConnectionInfo
    _consume_queue: asyncio.Queue[MQMessage]
    _subscribe_queue: asyncio.Queue[MQMessage]
    _stream_key: str
    _group_name: str
    _consumer_id: str
    _closed: bool
    # loop tasks for consuming messages
    _auto_claim_loop_task: asyncio.Task
    _read_messages_task: asyncio.Task
    _read_broadcast_messages_task: asyncio.Task

    def __init__(self, conn: RedisConnectionInfo, args: RedisMQArgs) -> None:
        self._conn = conn
        self._consume_queue = asyncio.Queue()
        self._subscribe_queue = asyncio.Queue()
        self._stream_key = args.stream_key
        self._group_name = args.group_name
        self._consumer_id = _generate_consumer_id(args.node_id)
        self._closed = False
        start_id = args.autoclaim_start_id or "0-0"
        self._auto_claim_loop_task = asyncio.create_task(
            self._auto_claim_loop(start_id, args.autoclaim_idle_timeout)
        )
        self._read_messages_task = asyncio.create_task(self._read_messages_loop())
        self._read_broadcast_messages_task = asyncio.create_task(
            self._read_broadcast_messages_loop()
        )

    async def send(self, payload: dict[bytes, bytes]) -> None:
        """
        Send a message to the queue.
        If the queue is full, the oldest message will be removed.
        The new message will be added to the end of the queue.
        """
        if self._closed:
            raise RuntimeError("Queue is closed")
        await self._conn.client.xadd(self._stream_key, payload, maxlen=_DEFAULT_QUEUE_MAX_LEN)

    async def consume_queue(self) -> AsyncGenerator[MQMessage, None]:  # type: ignore
        """
        Consume messages from the queue.
        This method will block until a message is available.

        This is a normal queue, so the message will be delivered to one consumer.
        Messages are consumed only once by one consumer.

        Consumer should call `done` method to acknowledge the message when it is processed.
        If the consumer does not call `done`, the message will be re-delivered after the
        `autoclaim_idle_timeout` period.
        """
        while not self._closed:
            try:
                yield await self._consume_queue.get()
            except asyncio.CancelledError:
                break

    async def subscribe_queue(self) -> AsyncGenerator[MQMessage, None]:  # type: ignore
        while not self._closed:
            try:
                yield await self._subscribe_queue.get()
            except asyncio.CancelledError:
                break

    async def done(self, msg_id: MessageId) -> None:
        await self._conn.client.xack(self._stream_key, self._group_name, msg_id)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self._conn.close()
        self._auto_claim_loop_task.cancel()
        self._read_messages_task.cancel()
        self._read_broadcast_messages_task.cancel()

    async def _auto_claim_loop(self, autoclaim_start_id: str, autoclaim_idle_timeout: int) -> None:
        log.debug("Starting auto claim loop for stream {}", self._stream_key)
        while not self._closed:
            try:
                next_start_id, claimed = await self._auto_claim(
                    autoclaim_start_id, autoclaim_idle_timeout
                )
                if not claimed:
                    await asyncio.sleep(_DEFAULT_AUTOCLAIM_INTERVAL / 1000)
                    continue
                autoclaim_start_id = next_start_id
            except redis.exceptions.ResponseError as e:
                await self._failover_consumer(e)
            except Exception as e:
                log.error("Error while auto claiming messages: {}", e)

    async def _auto_claim(
        self, autoclaim_start_id: str, autoclaim_idle_timeout: int
    ) -> tuple[str, bool]:
        reply = await redis_helper.execute(
            self._conn,
            lambda r: r.xautoclaim(
                self._stream_key,
                self._group_name,
                self._consumer_id,
                min_idle_time=autoclaim_idle_timeout,
                start_id=autoclaim_start_id,
                count=_DEFAULT_AUTOCLAIM_COUNT,
            ),
            command_timeout=autoclaim_idle_timeout / 1000,
        )
        if reply[0] == b"0-0":
            return autoclaim_start_id, False
        autoclaim_start_id = reply[0]
        for msg_id, msg_data in reply[1]:
            msg = MQMessage(msg_id, msg_data)
            if msg.retry():
                await self._retry_message(msg)
            else:
                # discard the message
                await self.done(msg_id)

        return autoclaim_start_id, True

    async def _retry_message(self, message: MQMessage) -> None:
        pipe = self._conn.client.pipeline(transaction=True)
        pipe.xack(self._stream_key, self._group_name, message.msg_id)
        pipe.xadd(self._stream_key, message.payload, maxlen=_DEFAULT_QUEUE_MAX_LEN)
        await pipe.execute()

    async def _read_messages_loop(self) -> None:
        log.debug("Reading messages from stream {}", self._stream_key)
        while not self._closed:
            try:
                await self._read_messages()
            except redis.exceptions.ResponseError as e:
                await self._failover_consumer(e)
            except Exception as e:
                log.error("Error while reading messages: {}", e)

    async def _read_messages(self) -> None:
        reply = await redis_helper.execute(
            self._conn,
            lambda r: r.xreadgroup(
                self._group_name,
                self._consumer_id,
                {self._stream_key: ">"},
                count=1,
                block=30_000,
            ),
        )
        if not reply:
            log.debug("No messages to read")
            return
        for _, events in reply:
            for msg_id, msg_data in events:
                if msg_data is None:
                    continue
                msg = MQMessage(msg_id, msg_data)
                await self._consume_queue.put(msg)

    async def _read_broadcast_messages_loop(self) -> None:
        log.debug("Reading broadcast messages from stream {}", self._stream_key)
        last_msg_id = "$"
        while not self._closed:
            try:
                last_msg_id = await self._read_broadcast_messages(last_msg_id)
            except redis.exceptions.ResponseError as e:
                await self._failover_consumer(e)
                last_msg_id = "$"
            except Exception as e:
                log.error("Error while reading broadcast messages: {}", e)

    async def _read_broadcast_messages(self, last_msg_id: str) -> str:
        reply = await redis_helper.execute(
            self._conn,
            lambda r: r.xread(
                {self._stream_key: last_msg_id},
                count=1,
                block=30_000,
            ),
        )
        if not reply:
            log.debug("No broadcast messages to read")
            return last_msg_id
        for _, events in reply:
            for msg_id, msg_data in events:
                if msg_data is None:
                    continue
                msg = MQMessage(msg_id, msg_data)
                await self._subscribe_queue.put(msg)
                last_msg_id = msg_id
        return last_msg_id

    async def _failover_consumer(self, e: redis.exceptions.ResponseError) -> None:
        # If the group does not exist, create it
        # and start the auto claim loop again
        if "NOGROUP" in str(e):
            log.warning(
                "Consumer group does not exist. Creating group {} for stream {}",
                self._group_name,
                self._stream_key,
            )
            try:
                await redis_helper.execute(
                    self._conn,
                    lambda r: r.xgroup_create(self._stream_key, self._group_name, mkstream=True),
                )
            except Exception as internal_exception:
                log.warning(
                    "Error while creating consumer group {} for stream {}: {}",
                    self._group_name,
                    self._stream_key,
                    internal_exception,
                )
        else:
            log.error("Error while reading messages: {}", e)


def _generate_consumer_id(node_id: Optional[str]) -> str:
    h = hashlib.sha1()
    h.update(str(node_id or socket.getfqdn()).encode("utf8"))
    hostname_hash = h.hexdigest()
    h = hashlib.sha1()
    h.update(__file__.encode("utf8"))
    installation_path_hash = h.hexdigest()
    pidx = process_index.get(0)
    return f"{hostname_hash}:{installation_path_hash}:{pidx}"



import uuid
from typing import Optional

from pydantic import Field

from ...api_handlers import BaseRequestModel


class ListGroupQuery(BaseRequestModel):
    group_id: Optional[uuid.UUID] = Field(default=None, alias="groupId")



import uuid
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Iterator, Optional

_request_id_var: ContextVar[str] = ContextVar("request_id")


def current_request_id() -> Optional[str]:
    """
    Get the current request ID from the context.
    Returns None if not set.
    """
    try:
        return _request_id_var.get()
    except LookupError:
        return None


@contextmanager
def with_request_id(request_id: Optional[str] = None) -> Iterator[None]:
    """
    context manager to set up the request ID.
    If request_id is not provided, generates a new UUID.
    This function returns the token that can be used to reset the context later.
    """

    if request_id is None:
        request_id = str(uuid.uuid4())

    token = _request_id_var.set(request_id)
    try:
        yield
    finally:
        # Reset the context variable to its previous state
        _request_id_var.reset(token)



import json
import logging
import os
import socket
import sys
from ipaddress import _BaseAddress as BaseIPAddress
from ipaddress import _BaseNetwork as BaseIPNetwork
from ipaddress import ip_address
from pathlib import Path, PosixPath
from typing import Awaitable, Callable, Iterable, Optional

import aiodns
import ifaddr
import psutil

from .utils import curl

__all__ = (
    "detect_cloud",
    "current_provider",
    "get_instance_id",
    "get_instance_ip",
    "get_instance_type",
    "get_instance_region",
    "get_root_fs_type",
    "get_wsl_version",
)

log = logging.getLogger(__spec__.name)


def is_containerized() -> bool:
    """
    Check if I am running inside a Linux container.
    """
    try:
        cginfo = Path("/proc/self/cgroup").read_text()
        if "/docker/" in cginfo or "/lxc/" in cginfo:
            return True
        return False
    except IOError:
        return False


def detect_cloud() -> Optional[str]:
    """
    Detect the cloud provider where I am running on.
    """
    # NOTE: Contributions are welcome!
    # Please add other cloud providers such as Rackspace, IBM BlueMix, etc.
    if sys.platform.startswith("linux"):
        # Google Cloud Platform or Amazon AWS (hvm)
        try:
            # AWS Nitro-based instances
            mb = Path("/sys/devices/virtual/dmi/id/board_vendor").read_text().lower()
            if "amazon" in mb:
                return "amazon"
        except IOError:
            pass
        try:
            bios = Path("/sys/devices/virtual/dmi/id/bios_version").read_text().lower()
            if "google" in bios:
                return "google"
            if "amazon" in bios:
                return "amazon"
        except IOError:
            pass
        # Microsoft Azure
        # https://gallery.technet.microsoft.com/scriptcenter/Detect-Windows-Azure-aed06d51
        # TODO: this only works with Debian/Ubuntu instances.
        # TODO: this does not work inside containers.
        try:
            dhcp = Path("/var/lib/dhcp/dhclient.eth0.leases").read_text()
            if "unknown-245" in dhcp:
                return "azure"
            # alternative method is to read /var/lib/waagent/GoalState.1.xml
            # but it requires sudo privilege.
        except IOError:
            pass
    return None


def fetch_local_ipaddrs(cidr: BaseIPNetwork) -> Iterable[BaseIPAddress]:
    proto = socket.AF_INET if cidr.version == 4 else socket.AF_INET6
    for adapter in ifaddr.get_adapters():
        if not adapter.ips:
            continue
        for entry in adapter.ips:
            if entry.is_IPv4 and proto == socket.AF_INET:
                assert isinstance(entry.ip, str)
                addr = ip_address(entry.ip)
            elif entry.is_IPv6 and proto == socket.AF_INET6:
                assert isinstance(entry.ip, tuple)
                addr = ip_address(entry.ip[0])
            else:
                continue
            if addr in cidr:
                yield addr


def get_root_fs_type() -> tuple[PosixPath, str]:
    for partition in psutil.disk_partitions():
        if partition.mountpoint == "/":
            return PosixPath(partition.device), partition.fstype
    raise RuntimeError("Could not find the root filesystem from the mounts.")


def get_wsl_version() -> int:
    """
    Returns the current WSL version we are running on, and 0 if we are not on WSL.

    ref) https://github.com/snapcore/snapd/blob/3a88dc38ca122eba97192dba3aad30f3bd3e3081/release/release.go#L116-L172
    """
    if not Path("/proc/sys/fs/binfmt_misc/WSLInterop").exists() and not Path("/run/WSL").exists():
        return 0
    try:
        _, root_fs_type = get_root_fs_type()
    except RuntimeError:
        return 2
    if root_fs_type in ("wslfs", "lxfs"):
        return 1
    return 2


# Detect upon module load.
current_provider = detect_cloud()
if current_provider is None:
    log.info("Detected environment: on-premise setup")
    log.info("The agent node ID is set using the hostname.")
else:
    log.info(f"Detected environment: {current_provider} cloud")
    log.info("The agent node ID will follow the instance ID.")

_defined: bool = False
get_instance_id: Callable[[], Awaitable[str]]
get_instance_ip: Callable[[Optional[BaseIPNetwork]], Awaitable[str]]
get_instance_type: Callable[[], Awaitable[str]]
get_instance_region: Callable[[], Awaitable[str]]


def _define_functions():
    global _defined
    global get_instance_id
    global get_instance_ip
    global get_instance_type
    global get_instance_region
    if _defined:
        return

    if current_provider == "amazon":
        # ref: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
        _metadata_prefix = "http://169.254.169.254/latest/meta-data/"
        _dynamic_prefix = "http://169.254.169.254/latest/dynamic/"

        async def _get_instance_id() -> str:
            return await curl(_metadata_prefix + "instance-id", lambda: f"i-{socket.gethostname()}")

        async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
            return await curl(_metadata_prefix + "local-ipv4", "127.0.0.1")

        async def _get_instance_type() -> str:
            return await curl(_metadata_prefix + "instance-type", "unknown")

        async def _get_instance_region() -> str:
            doc = await curl(_dynamic_prefix + "instance-identity/document", None)
            if doc is None:
                return "amazon/unknown"
            region = json.loads(doc)["region"]
            return f"amazon/{region}"

    elif current_provider == "azure":
        # ref: https://docs.microsoft.com/azure/virtual-machines/virtual-machines-instancemetadataservice-overview
        _metadata_prefix = "http://169.254.169.254/metadata/instance"

        async def _get_instance_id() -> str:
            data = await curl(
                _metadata_prefix,
                None,
                params={"version": "2017-03-01"},
                headers={"Metadata": "true"},
            )
            if data is None:
                return f"i-{socket.gethostname()}"
            o = json.loads(data)
            return o["compute"]["vmId"]

        async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
            data = await curl(
                _metadata_prefix,
                None,
                params={"version": "2017-03-01"},
                headers={"Metadata": "true"},
            )
            if data is None:
                return "127.0.0.1"
            o = json.loads(data)
            return o["network"]["interface"][0]["ipv4"]["ipaddress"][0]["ipaddress"]

        async def _get_instance_type() -> str:
            data = await curl(
                _metadata_prefix,
                None,
                params={"version": "2017-03-01"},
                headers={"Metadata": "true"},
            )
            if data is None:
                return "unknown"
            o = json.loads(data)
            return o["compute"]["vmSize"]

        async def _get_instance_region() -> str:
            data = await curl(
                _metadata_prefix,
                None,
                params={"version": "2017-03-01"},
                headers={"Metadata": "true"},
            )
            if data is None:
                return "azure/unknown"
            o = json.loads(data)
            region = o["compute"]["location"]
            return f"azure/{region}"

    elif current_provider == "google":
        # ref: https://cloud.google.com/compute/docs/storing-retrieving-metadata
        _metadata_prefix = "http://metadata.google.internal/computeMetadata/v1/"

        async def _get_instance_id() -> str:
            return await curl(
                _metadata_prefix + "instance/id",
                lambda: f"i-{socket.gethostname()}",
                headers={"Metadata-Flavor": "Google"},
            )

        async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
            return await curl(
                _metadata_prefix + "instance/network-interfaces/0/ip",
                "127.0.0.1",
                headers={"Metadata-Flavor": "Google"},
            )

        async def _get_instance_type() -> str:
            return await curl(
                _metadata_prefix + "instance/machine-type",
                "unknown",
                headers={"Metadata-Flavor": "Google"},
            )

        async def _get_instance_region() -> str:
            zone = await curl(
                _metadata_prefix + "instance/zone", "unknown", headers={"Metadata-Flavor": "Google"}
            )
            region = zone.rsplit("-", 1)[0]
            return f"google/{region}"

    else:
        _metadata_prefix = None

        async def _get_instance_id() -> str:
            return f"i-{socket.gethostname()}"

        async def _get_instance_ip(subnet_hint: Optional[BaseIPNetwork] = None) -> str:
            if subnet_hint is not None and subnet_hint.prefixlen > 0:
                local_ipaddrs = [*fetch_local_ipaddrs(subnet_hint)]
                if local_ipaddrs:
                    return str(local_ipaddrs[0])
                raise RuntimeError("Could not find my IP address bound to subnet {}", subnet_hint)
            try:
                myself = socket.gethostname()
                resolver = aiodns.DNSResolver()
                result = await resolver.gethostbyname(myself, socket.AF_INET)
                return result.addresses[0]
            except aiodns.error.DNSError:
                return "127.0.0.1"

        async def _get_instance_type() -> str:
            return "default"

        async def _get_instance_region() -> str:
            return os.environ.get("BACKEND_REGION", "local")

    get_instance_id = _get_instance_id
    get_instance_ip = _get_instance_ip
    get_instance_type = _get_instance_type
    get_instance_region = _get_instance_region
    _defined = True


_define_functions()



from __future__ import annotations

import asyncio
import logging
import time
import uuid
import weakref
from dataclasses import dataclass
from typing import (
    Awaitable,
    Callable,
    Concatenate,
    Final,
    Mapping,
    Optional,
    Protocol,
    Self,
    TypeAlias,
    Union,
)

from redis.asyncio import Redis
from redis.asyncio.client import Pipeline

from ai.backend.common.bgtask.types import BgtaskStatus
from ai.backend.common.exception import (
    BackendAIError,
    BgtaskNotFoundError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.logging import BraceStyleAdapter

from .. import redis_helper
from ..events.bgtask import (
    BaseBgtaskDoneEvent,
    BgtaskCancelledEvent,
    BgtaskDoneEvent,
    BgtaskFailedEvent,
    BgtaskPartialSuccessEvent,
    BgtaskUpdatedEvent,
)
from ..events.dispatcher import (
    EventProducer,
)
from ..types import DispatchResult, RedisConnectionInfo, Sentinel

sentinel: Final = Sentinel.TOKEN
log = BraceStyleAdapter(logging.getLogger(__spec__.name))


BgtaskEvents: TypeAlias = (
    BgtaskUpdatedEvent
    | BgtaskDoneEvent
    | BgtaskCancelledEvent
    | BgtaskFailedEvent
    | BgtaskPartialSuccessEvent
)

_MAX_BGTASK_ARCHIVE_PERIOD: Final = 86400  # 24  hours


@dataclass
class BgTaskInfo:
    status: BgtaskStatus
    msg: str
    started_at: str
    last_update: str
    current: str = "0"
    total: str = "0"

    @classmethod
    def started(cls, msg: str = "") -> Self:
        now = str(time.time())
        return cls(
            status=BgtaskStatus.STARTED,
            msg=msg,
            started_at=now,
            last_update=now,
            current="0",
            total="0",
        )

    @classmethod
    def finished(cls, status: BgtaskStatus, msg: str = "") -> Self:
        now = str(time.time())
        return cls(
            status=status,
            msg=msg,
            started_at="0",
            last_update=now,
            current="0",
            total="0",
        )

    def to_dict(self) -> Mapping[str | bytes, str]:
        return {
            "status": str(self.status),
            "msg": self.msg,
            "started_at": self.started_at,
            "last_update": self.last_update,
            "current": self.current,
            "total": self.total,
        }


class ProgressReporter:
    total_progress: Union[int, float]
    current_progress: Union[int, float]

    _event_producer: Final[EventProducer]
    _redis_client: RedisConnectionInfo
    _task_id: Final[uuid.UUID]

    def __init__(
        self,
        redis_client: RedisConnectionInfo,
        event_producer: EventProducer,
        task_id: uuid.UUID,
        current_progress: int = 0,
        total_progress: int = 0,
    ) -> None:
        self._redis_client = redis_client
        self._event_producer = event_producer
        self._task_id = task_id
        self.current_progress = current_progress
        self.total_progress = total_progress

    async def update(
        self,
        increment: Union[int, float] = 0,
        message: str | None = None,
    ) -> None:
        self.current_progress += increment
        # keep the state as local variables because they might be changed
        # due to interleaving at await statements below.
        current, total = self.current_progress, self.total_progress

        async def _pipe_builder(r: Redis) -> Pipeline:
            pipe = r.pipeline(transaction=False)
            tracker_key = f"bgtask.{self._task_id}"
            await pipe.hset(
                tracker_key,
                mapping={
                    "current": str(current),
                    "total": str(total),
                    "msg": message or "",
                    "last_update": str(time.time()),
                },
            )
            await pipe.expire(tracker_key, _MAX_BGTASK_ARCHIVE_PERIOD)
            return pipe

        await redis_helper.execute(self._redis_client, _pipe_builder)
        await self._event_producer.produce_event(
            BgtaskUpdatedEvent(
                self._task_id,
                message=message,
                current_progress=current,
                total_progress=total,
            ),
        )


BackgroundTask = Callable[
    Concatenate[ProgressReporter, ...], Awaitable[str | DispatchResult | None]
]


class BackgroundTaskObserver(Protocol):
    def observe_bgtask_started(self, *, task_name: str) -> None: ...
    def observe_bgtask_done(
        self, *, task_name: str, status: str, duration: float, error_code: Optional[ErrorCode]
    ) -> None: ...


class NopBackgroundTaskObserver:
    def observe_bgtask_started(self, *, task_name: str) -> None:
        pass

    def observe_bgtask_done(
        self, *, task_name: str, status: str, duration: float, error_code: Optional[ErrorCode]
    ) -> None:
        pass


class BackgroundTaskManager:
    _redis_client: RedisConnectionInfo
    _event_producer: EventProducer
    _ongoing_tasks: weakref.WeakSet[asyncio.Task]
    _metric_observer: BackgroundTaskObserver
    _dict_lock: asyncio.Lock

    def __init__(
        self,
        redis_client: RedisConnectionInfo,
        event_producer: EventProducer,
        *,
        bgtask_observer: BackgroundTaskObserver = NopBackgroundTaskObserver(),
    ) -> None:
        self._redis_client = redis_client
        self._event_producer = event_producer
        self._ongoing_tasks = weakref.WeakSet()
        self._metric_observer = bgtask_observer
        self._dict_lock = asyncio.Lock()

    async def fetch_bgtask_info(
        self,
        task_id: uuid.UUID,
    ) -> BgTaskInfo:
        tracker_key = _tracker_id(task_id)
        task_info_dict = await redis_helper.execute(
            self._redis_client,
            lambda r: r.hgetall(tracker_key),
            encoding="utf-8",
        )
        if task_info_dict is None:
            # The task ID is invalid or represents a task completed more than timeout.
            raise BgtaskNotFoundError("No such background task.")

        task_info_dict["status"] = BgtaskStatus(task_info_dict["status"])
        task_info = BgTaskInfo(**task_info_dict)
        return task_info

    async def start(
        self,
        func: BackgroundTask,
        name: Optional[str] = None,
        **kwargs,
    ) -> uuid.UUID:
        task_id = uuid.uuid4()
        await self._update_bgtask_status(task_id=task_id, status=BgtaskStatus.STARTED, msg="")
        task = asyncio.create_task(self._wrapper_task(func, task_id, name, **kwargs))
        self._ongoing_tasks.add(task)
        return task_id

    async def shutdown(self) -> None:
        log.info("Cancelling remaining background tasks...")
        for task in self._ongoing_tasks.copy():
            if task.done():
                continue
            try:
                task.cancel()
                await task
            except asyncio.CancelledError:
                pass

    async def _update_bgtask_status(
        self,
        task_id: uuid.UUID,
        status: BgtaskStatus,
        msg: str = "",
    ) -> None:
        tracker_key = _tracker_id(task_id)

        async def _pipe_builder(r: Redis) -> Pipeline:
            pipe = r.pipeline()
            task_info: BgTaskInfo
            if status.finished():
                task_info = BgTaskInfo.finished(status=status, msg=msg)
            else:
                task_info = BgTaskInfo.started(msg=msg)
            mapping = task_info.to_dict()
            pipe.hset(tracker_key, mapping=mapping)
            pipe.expire(tracker_key, _MAX_BGTASK_ARCHIVE_PERIOD)
            return pipe

        await redis_helper.execute(self._redis_client, _pipe_builder)

    def _convert_bgtask_to_event(
        self, task_id: uuid.UUID, bgtask_result: DispatchResult | str | None
    ) -> BaseBgtaskDoneEvent:
        # legacy
        if bgtask_result is None or isinstance(bgtask_result, str):
            return BgtaskDoneEvent(task_id, bgtask_result)

        message = bgtask_result.message()
        if bgtask_result.has_error():
            return BgtaskPartialSuccessEvent(
                task_id=task_id, message=message, errors=bgtask_result.errors
            )
        else:
            return BgtaskDoneEvent(task_id=task_id, message=message)

    async def _run_bgtask(
        self,
        func: BackgroundTask,
        task_id: uuid.UUID,
        **kwargs,
    ) -> BaseBgtaskDoneEvent:
        reporter = ProgressReporter(self._redis_client, self._event_producer, task_id)
        bgtask_result = await func(reporter, **kwargs)
        return self._convert_bgtask_to_event(task_id, bgtask_result)

    async def _observe_bgtask(
        self,
        func: BackgroundTask,
        task_id: uuid.UUID,
        task_name: Optional[str],
        **kwargs,
    ) -> BaseBgtaskDoneEvent:
        self._metric_observer.observe_bgtask_started(task_name=task_name or func.__name__)
        start_time = time.perf_counter()
        task_name = task_name or func.__name__
        status = BgtaskStatus.STARTED
        error_code: Optional[ErrorCode] = None
        msg = "no message"
        try:
            bgtask_result_event = await self._run_bgtask(func, task_id, **kwargs)
            status = bgtask_result_event.status()
            msg = bgtask_result_event.message or msg
        except asyncio.CancelledError:
            status = BgtaskStatus.CANCELLED
            error_code = ErrorCode(
                domain=ErrorDomain.BGTASK,
                operation=ErrorOperation.EXECUTE,
                error_detail=ErrorDetail.CANCELED,
            )
            return BgtaskCancelledEvent(task_id, "")
        except BackendAIError as e:
            status = BgtaskStatus.FAILED
            error_code = e.error_code()
            log.error("Task {} ({}): BackendAIError: {}", task_id, task_name, e)
            return BgtaskFailedEvent(task_id, repr(e))
        except Exception as e:
            status = BgtaskStatus.FAILED
            error_code = ErrorCode(
                domain=ErrorDomain.BGTASK,
                operation=ErrorOperation.EXECUTE,
                error_detail=ErrorDetail.INTERNAL_ERROR,
            )
            log.error("Task {} ({}): unhandled error", task_id, task_name)
            return BgtaskFailedEvent(task_id, repr(e))
        finally:
            duration = time.perf_counter() - start_time
            self._metric_observer.observe_bgtask_done(
                task_name=task_name,
                status=status,
                duration=duration,
                error_code=error_code,
            )
            await self._update_bgtask_status(task_id, status, msg=msg)

        return bgtask_result_event

    async def _wrapper_task(
        self,
        func: BackgroundTask,
        task_id: uuid.UUID,
        task_name: Optional[str],
        **kwargs,
    ) -> None:
        bgtask_result_event = await self._observe_bgtask(func, task_id, task_name, **kwargs)
        await self._event_producer.produce_event(bgtask_result_event)
        log.info(
            "Task {} ({}): {}", task_id, task_name or "", bgtask_result_event.__class__.__name__
        )


def _tracker_id(task_id: uuid.UUID) -> str:
    return f"bgtask.{task_id}"



from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Optional

from ai.backend.logging import BraceStyleAdapter

from ..events.types import AbstractEvent
from ..types import AgentId
from . import AbstractPlugin, BasePluginContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AbstractEventDispatcherPlugin(AbstractPlugin, metaclass=ABCMeta):
    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    @abstractmethod
    async def handle_event(
        self,
        source: AgentId,
        event: AbstractEvent,
    ) -> None:
        raise NotImplementedError


class EventDispatcherPluginContext(BasePluginContext[AbstractEventDispatcherPlugin]):
    plugin_group = "backendai_event_dispatcher_v20"

    async def handle_event(
        self,
        context: None,
        source: AgentId,
        event: AbstractEvent,
    ) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.handle_event(source, event)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.events.types import AbstractEvent, EventDomain
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.types import KernelId, ModelServiceStatus, SessionId


@dataclass
class ModelServiceStatusEventArgs(AbstractEvent):
    kernel_id: KernelId
    session_id: SessionId
    model_name: str
    new_status: ModelServiceStatus

    def serialize(self) -> tuple:
        return (
            str(self.kernel_id),
            str(self.session_id),
            self.model_name,
            self.new_status.value,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            kernel_id=KernelId(uuid.UUID(value[0])),
            session_id=SessionId(uuid.UUID(value[1])),
            model_name=value[2],
            new_status=ModelServiceStatus(value[3]),
        )

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.MODEL_SERVING

    @override
    def domain_id(self) -> Optional[str]:
        return None

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


class ModelServiceStatusEvent(ModelServiceStatusEventArgs):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "model_service_status_updated"


@dataclass
class RouteCreationEvent(AbstractEvent):
    route_id: uuid.UUID

    def serialize(self) -> tuple:
        return (str(self.route_id),)

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(uuid.UUID(value[0]))

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.MODEL_ROUTE

    @override
    def domain_id(self) -> Optional[str]:
        return str(self.route_id)

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


class RouteCreatedEvent(RouteCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "route_created"



from dataclasses import asdict, dataclass
from typing import Any, Mapping, Optional, override

from .user_event import UserEvent


@dataclass
class UserBgTaskEvent(UserEvent):  # TODO: Need to separate UserPartialSuccessEvent
    task_id: str
    message: str
    status: str
    current_progress: str
    total_progress: str

    @override
    def event_name(self) -> Optional[str]:
        return self.status

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass
class UserBgtaskUpdatedEvent(UserEvent):
    task_id: str
    message: str
    current_progress: float
    total_progress: float

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_updated"

    @override
    def retry_count(self) -> Optional[int]:
        return 5

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass
class UserBgtaskDoneEvent(UserEvent):
    task_id: str
    message: str

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_done"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass
class UserBgtaskPartialSuccessEvent(UserEvent):
    task_id: str
    message: str
    errors: list[str]

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_done"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass
class UserBgtaskCancelledEvent(UserEvent):
    task_id: str
    message: str

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_cancelled"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass
class UserBgtaskFailedEvent(UserEvent):
    task_id: str
    message: str

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_failed"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)


@dataclass
class UserBgtaskAlreadyDoneEvent(UserEvent):
    task_id: str
    message: str
    current_progress: float
    total_progress: float

    @override
    def event_name(self) -> Optional[str]:
        return "bgtask_already_done"

    @override
    def retry_count(self) -> Optional[int]:
        return None

    @override
    def user_event_mapping(self) -> Mapping[str, Any]:
        return asdict(self)



from dataclasses import dataclass
from typing import Optional, Self, override

from ai.backend.common.docker import ImageRef
from ai.backend.common.events.types import AbstractEvent, EventDomain
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.types import AgentId


@dataclass
class BaseImageEvent(AbstractEvent):
    image: str
    agent_id: AgentId

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.IMAGE

    @override
    def domain_id(self) -> Optional[str]:
        return self.image

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class ImagePullStartedEvent(BaseImageEvent):
    timestamp: float
    image_ref: Optional[ImageRef] = None

    @override
    def serialize(self) -> tuple:
        if self.image_ref is None:
            return (self.image, str(self.agent_id), self.timestamp)

        return (
            self.image,
            str(self.agent_id),
            self.timestamp,
            self.image_ref,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple) -> Self:
        # Backward compatibility
        if len(value) <= 3:
            return cls(
                image=value[0],
                agent_id=AgentId(value[1]),
                timestamp=value[2],
            )

        return cls(
            image=value[0],
            agent_id=AgentId(value[1]),
            timestamp=value[2],
            image_ref=value[3],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "image_pull_started"


@dataclass
class ImagePullFinishedEvent(BaseImageEvent):
    timestamp: float
    msg: Optional[str] = None
    image_ref: Optional[ImageRef] = None

    @override
    def serialize(self) -> tuple:
        return (
            self.image,
            str(self.agent_id),
            self.timestamp,
            self.msg,
            self.image_ref,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple) -> Self:
        # Backward compatibility
        if len(value) <= 4:
            return cls(
                image=value[0],
                agent_id=AgentId(value[1]),
                timestamp=value[2],
                msg=value[3],
            )

        return cls(
            image=value[0],
            agent_id=AgentId(value[1]),
            timestamp=value[2],
            msg=value[3],
            image_ref=value[4],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "image_pull_finished"


@dataclass
class ImagePullFailedEvent(BaseImageEvent):
    msg: str
    image_ref: Optional[ImageRef] = None

    @override
    def serialize(self) -> tuple:
        if self.image_ref is None:
            return (self.image, str(self.agent_id), self.msg)
        return (self.image, str(self.agent_id), self.msg, self.image_ref)

    @classmethod
    @override
    def deserialize(cls, value: tuple) -> Self:
        # Backward compatibility
        if len(value) <= 3:
            return cls(
                image=value[0],
                agent_id=AgentId(value[1]),
                msg=value[2],
            )

        return cls(
            image=value[0],
            agent_id=AgentId(value[1]),
            msg=value[2],
            image_ref=value[3],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "image_pull_failed"



from typing import Optional, Self, override

from ai.backend.common.events.types import AbstractEvent, EventDomain
from ai.backend.common.events.user_event.user_event import UserEvent


class BaseScheduleEvent(AbstractEvent):
    @override
    def serialize(self) -> tuple:
        return tuple()

    @classmethod
    @override
    def deserialize(cls, value: tuple) -> Self:
        return cls()

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.SCHEDULE

    @override
    def domain_id(self) -> Optional[str]:
        return None

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


class DoScheduleEvent(BaseScheduleEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "do_schedule"


class DoCheckPrecondEvent(BaseScheduleEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "do_check_precond"


class DoStartSessionEvent(BaseScheduleEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "do_start_session"


class DoScaleEvent(BaseScheduleEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "do_scale"



import asyncio
import logging
import time
import uuid
from abc import ABC, abstractmethod
from typing import Any, Self, Sequence

from pydantic import BaseModel, Field

from ai.backend.logging.utils import BraceStyleAdapter

_DEFAULT_HEARTBEAT_TIMEOUT = 60 * 5  # 5 minutes
_DEFAULT_SWEEP_INTERVAL = 60 * 10  # 10 minutes

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ServiceEndpoint(BaseModel):
    """
    Service endpoint.
    """

    address: str
    port: int
    protocol: str
    prometheus_address: str


class HealthStatus(BaseModel):
    """
    Health status of a service.
    """

    registration_time: float = Field(default_factory=time.time)
    last_heartbeat: float = Field(default_factory=time.time)

    def check_healthy(self, timeout: float = _DEFAULT_HEARTBEAT_TIMEOUT) -> bool:
        """
        Check if the service is healthy.
        :return: True if the service is healthy, False otherwise.
        """
        now = time.time()
        return (now - self.last_heartbeat) < timeout

    def update_heartbeat(self) -> None:
        """
        Update the last heartbeat time.
        """
        self.last_heartbeat = time.time()

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, HealthStatus):
            return False
        # Heartbeat time is not considered for equality check
        return self.registration_time == value.registration_time


class ServiceMetadata(BaseModel):
    """
    Metadata for a service.
    """

    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    display_name: str
    service_group: str
    version: str
    endpoint: ServiceEndpoint
    health_status: HealthStatus = Field(default_factory=HealthStatus)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the service metadata to a dictionary.
        :return: Dictionary representation of the service metadata.
        """
        return self.model_dump()


class ServiceDiscovery(ABC):
    """
    Abstract class for service discovery.
    This class is used to discover services in a distributed system.
    """

    @abstractmethod
    async def register(self, service: ServiceMetadata) -> None:
        """
        Register a service.
        :param service: Service metadata.
        """
        raise NotImplementedError

    @abstractmethod
    async def unregister(self, service_group: str, service_id: uuid.UUID) -> None:
        """
        Unregister a service.
        :param service_group: Name of the service group.
        :param service_id: UUID of the service.
        """
        raise NotImplementedError

    @abstractmethod
    async def heartbeat(self, service_group: str, service_id: uuid.UUID) -> None:
        """
        Send a heartbeat to the service discovery.
        :param service_group: Name of the service group.
        :param service_id: UUID of the service.
        """
        raise NotImplementedError

    @abstractmethod
    async def discover(self) -> Sequence[ServiceMetadata]:
        """
        Discover services.
        :return: List of service metadata.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_service_group(self, service_group: str) -> Sequence[ServiceMetadata]:
        """
        Get services by group.
        :param service_group: Name of the service group.
        :return: List of service metadata.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_service(self, service_group: str, service_id: uuid.UUID) -> ServiceMetadata:
        """
        Get service address by name.
        :param service_group: Name of the service group.
        :param service_id: UUID of the service.
        :return: Service metadata.
        """
        raise NotImplementedError


class ServiceDiscoveryLoop:
    """
    Service discovery loop.
    This class is used to discover services in a distributed system.
    """

    _service_discovery: ServiceDiscovery
    _metadata: ServiceMetadata
    _interval_seconds: int
    _closed: bool = False
    _run_service_task: asyncio.Task[None]
    _sweep_unhealthy_services_task: asyncio.Task[None]

    def __init__(
        self,
        service_discovery: ServiceDiscovery,
        metadata: ServiceMetadata,
        interval_seconds: int = 60,
    ) -> None:
        self._service_discovery = service_discovery
        self._metadata = metadata
        self._interval_seconds = interval_seconds
        self._closed = False
        self._run_service_task = asyncio.create_task(self._run_service_loop())
        self._sweep_unhealthy_services_task = asyncio.create_task(
            self._sweep_unhealthy_services_loop()
        )

    @property
    def metadata(self) -> ServiceMetadata:
        return self._metadata

    def close(self) -> None:
        """
        Close the service discovery loop.
        """
        if self._closed:
            return
        self._closed = True
        self._run_service_task.cancel()
        self._sweep_unhealthy_services_task.cancel()

    async def _sweep_unhealthy_services_loop(self) -> None:
        """
        Sweep unhealthy services.
        This method is used to sweep unhealthy services in the service discovery.
        """
        while not self._closed:
            try:
                services = await self._service_discovery.discover()
                for service in services:
                    if not service.health_status.check_healthy():
                        await self._service_discovery.unregister(
                            service_group=service.service_group,
                            service_id=service.id,
                        )
            except Exception as e:
                log.error("Error sweeping unhealthy services: {}", e)
            await asyncio.sleep(_DEFAULT_SWEEP_INTERVAL)

    async def _run_service_loop(self) -> None:
        log.info(
            "Registering service {} with ID {} in group {}",
            self._metadata.display_name,
            self._metadata.id,
            self._metadata.service_group,
        )
        await self._service_discovery.register(self._metadata)
        while not self._closed:
            try:
                await self._service_discovery.heartbeat(
                    service_group=self._metadata.service_group,
                    service_id=self._metadata.id,
                )
            except Exception as e:
                log.error("Error sending heartbeat: {}", e)
            await asyncio.sleep(self._interval_seconds)
        await self._service_discovery.unregister(
            service_group=self._metadata.service_group,
            service_id=self._metadata.id,
        )



import enum


class BgtaskStatus(enum.StrEnum):
    STARTED = "bgtask_started"
    DONE = "bgtask_done"
    CANCELLED = "bgtask_cancelled"
    FAILED = "bgtask_failed"
    PARTIAL_SUCCESS = "bgtask_partial_success"

    def finished(self) -> bool:
        return self in {self.DONE, self.CANCELLED, self.FAILED, self.PARTIAL_SUCCESS}



import asyncio
import hashlib
import logging
import socket
from dataclasses import dataclass
from typing import AsyncGenerator, Optional

import hiredis
from aiotools.server import process_index

from ai.backend.common.redis_client import RedisConnection
from ai.backend.logging.utils import BraceStyleAdapter

from ..types import RedisTarget
from .queue import AbstractMessageQueue, MessageId, MQMessage

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_DEFAULT_AUTOCLAIM_IDLE_TIMEOUT = 300_000  # 5 minutes
_DEFAULT_AUTOCLAIM_INTERVAL = 60_000
_DEFAULT_AUTOCLAIM_COUNT = 64
_DEFAULT_QUEUE_MAX_LEN = 128


def _make_pieces(payload: dict[bytes, bytes]) -> list[bytes]:
    pieces = []
    for k, v in payload.items():
        pieces.append(k)
        pieces.append(v)
    return pieces


@dataclass
class HiRedisMQArgs:
    # Required arguments
    stream_key: str
    group_name: str
    node_id: str
    db: int
    # Optional arguments
    autoclaim_idle_timeout: int = _DEFAULT_AUTOCLAIM_IDLE_TIMEOUT
    autoclaim_start_id: Optional[str] = None


class HiRedisQueue(AbstractMessageQueue):
    _target: RedisTarget
    _db: int
    _consume_queue: asyncio.Queue[MQMessage]
    _subscribe_queue: asyncio.Queue[MQMessage]
    _stream_key: str
    _group_name: str
    _consumer_id: str
    _closed: bool
    # loop tasks for consuming messages
    _auto_claim_loop_task: asyncio.Task
    _read_messages_task: asyncio.Task
    _read_broadcast_messages_task: asyncio.Task

    def __init__(self, target: RedisTarget, args: HiRedisMQArgs) -> None:
        self._target = target
        self._db = args.db
        self._consume_queue = asyncio.Queue()
        self._subscribe_queue = asyncio.Queue()
        self._stream_key = args.stream_key
        self._group_name = args.group_name
        self._consumer_id = _generate_consumer_id(args.node_id)
        self._closed = False
        start_id = args.autoclaim_start_id or "0-0"
        self._auto_claim_loop_task = asyncio.create_task(
            self._auto_claim_loop(start_id, args.autoclaim_idle_timeout)
        )
        self._read_messages_task = asyncio.create_task(self._read_messages_loop())
        self._read_broadcast_messages_task = asyncio.create_task(
            self._read_broadcast_messages_loop()
        )

    async def send(self, payload: dict[bytes, bytes]) -> None:
        async with RedisConnection(self._target, db=self._db) as client:
            pieces = _make_pieces(payload)
            await client.execute([
                "XADD",
                self._stream_key,
                "MAXLEN",
                _DEFAULT_QUEUE_MAX_LEN,
                "*",
                *pieces,
            ])

    async def consume_queue(self) -> AsyncGenerator[MQMessage, None]:  # type: ignore
        """
        Consume messages from the queue.
        This method will block until a message is available.

        This is a normal queue, so the message will be delivered to one consumer.
        Messages are consumed only once by one consumer.

        Consumer should call `done` method to acknowledge the message when it is processed.
        If the consumer does not call `done`, the message will be re-delivered after the
        `autoclaim_idle_timeout` period.
        """
        while not self._closed:
            try:
                yield await self._consume_queue.get()
            except asyncio.CancelledError:
                break

    async def subscribe_queue(self) -> AsyncGenerator[MQMessage, None]:  # type: ignore
        while not self._closed:
            try:
                yield await self._subscribe_queue.get()
            except asyncio.CancelledError:
                break

    async def done(self, msg_id: MessageId) -> None:
        async with RedisConnection(self._target, db=self._db) as client:
            await client.execute([
                "XACK",
                self._stream_key,
                self._group_name,
                msg_id,
            ])

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        self._auto_claim_loop_task.cancel()
        self._read_messages_task.cancel()
        self._read_broadcast_messages_task.cancel()

    async def _auto_claim_loop(self, autoclaim_start_id: str, autoclaim_idle_timeout: int) -> None:
        log.debug("Starting auto claim loop for stream {}", self._stream_key)
        while not self._closed:
            try:
                next_start_id, claimed = await self._auto_claim(
                    autoclaim_start_id, autoclaim_idle_timeout
                )
                if not claimed:
                    await asyncio.sleep(_DEFAULT_AUTOCLAIM_INTERVAL / 1000)
                    continue
                autoclaim_start_id = next_start_id
            except hiredis.HiredisError as e:
                await self._failover_consumer(e)
            except Exception as e:
                log.error("Error while auto claiming messages: {}", e)

    async def _auto_claim(
        self, autoclaim_start_id: str, autoclaim_idle_timeout: int
    ) -> tuple[str, bool]:
        async with RedisConnection(self._target, db=self._db) as client:
            reply = await client.execute(
                [
                    "XAUTOCLAIM",
                    self._stream_key,
                    self._group_name,
                    self._consumer_id,
                    str(autoclaim_idle_timeout),
                    autoclaim_start_id,
                    "COUNT",
                    _DEFAULT_AUTOCLAIM_COUNT,
                ],
                command_timeout=(autoclaim_idle_timeout + 5_000) / 1000,
            )
            if reply is None or reply[0] == b"0-0":
                return autoclaim_start_id, False
            for stream_msg in reply.values():
                for msg_id, messages in stream_msg:
                    payload = {}
                    for i in range(0, len(messages), 2):
                        key = messages[i]
                        value = messages[i + 1]
                        payload[key] = value
                    msg = MQMessage(msg_id, payload)
                    if msg.retry():
                        await self._retry_message(msg)
                    else:
                        # discard the message
                        await self.done(msg_id)
        return autoclaim_start_id, True

    async def _retry_message(self, message: MQMessage) -> None:
        pieces = _make_pieces(message.payload)
        async with RedisConnection(self._target, db=self._db) as client:
            await client.pipeline([
                [
                    "XACK",
                    self._stream_key,
                    self._group_name,
                    message.msg_id,
                ],
                [
                    "XADD",
                    self._stream_key,
                    "MAXLEN",
                    _DEFAULT_QUEUE_MAX_LEN,
                    "*",
                    *pieces,
                ],
            ])

    async def _read_messages_loop(self) -> None:
        log.debug("Reading messages from stream {}", self._stream_key)
        while not self._closed:
            try:
                await self._read_messages()
            except hiredis.HiredisError as e:
                await self._failover_consumer(e)
            except Exception as e:
                log.error("Error while reading messages: {}", e)

    async def _read_messages(self) -> None:
        log.debug("Reading messages from stream {}", self._stream_key)
        async with RedisConnection(self._target, db=self._db) as client:
            reply = await client.execute(
                [
                    "XREADGROUP",
                    "GROUP",
                    self._group_name,
                    self._consumer_id,
                    "COUNT",
                    1,
                    "BLOCK",
                    30_000,
                    "STREAMS",
                    self._stream_key,
                    ">",  # fetch messages not seen by other consumers
                ],
                command_timeout=5,
            )
            if reply is None:
                log.debug("No messages to read")
                return
            for stream_msg in reply.values():
                for msg_id, messages in stream_msg:
                    payload = {}
                    for i in range(0, len(messages), 2):
                        key = messages[i]
                        value = messages[i + 1]
                        payload[key] = value
                    msg = MQMessage(msg_id, payload)
                    await self._consume_queue.put(msg)

    async def _read_broadcast_messages_loop(self) -> None:
        log.debug("Reading broadcast messages from stream {}", self._stream_key)
        last_msg_id = b"$"
        while not self._closed:
            try:
                last_msg_id = await self._read_broadcast_messages(last_msg_id)
            except hiredis.HiredisError as e:
                await self._failover_consumer(e)
                last_msg_id = b"$"
            except Exception as e:
                log.error("Error while reading broadcast messages: {}", e)

    async def _read_broadcast_messages(self, last_msg_id: bytes) -> bytes:
        async with RedisConnection(self._target, db=self._db) as client:
            reply = await client.execute(
                ["XREAD", "BLOCK", "30000", "STREAMS", self._stream_key, last_msg_id],
                command_timeout=5,
            )
            if reply is None:
                log.debug("No messages to read")
                return last_msg_id
            for stream_msg in reply.values():
                for msg_id, messages in stream_msg:
                    payload = {}
                    for i in range(0, len(messages), 2):
                        key = messages[i]
                        value = messages[i + 1]
                        payload[key] = value
                    msg = MQMessage(msg_id, payload)
                    await self._subscribe_queue.put(msg)
                    last_msg_id = msg_id
        return last_msg_id

    async def _failover_consumer(self, e: hiredis.HiredisError) -> None:
        # If the group does not exist, create it
        # and start the auto claim loop again
        if not e.args[0].startswith("NOGROUP "):
            log.error("Error while auto claiming messages: {}", e)
            return
        try:
            async with RedisConnection(self._target, db=self._db) as client:
                await client.execute([
                    "XGROUP",
                    "CREATE",
                    self._stream_key,
                    self._group_name,
                    "$",
                    "MKSTREAM",
                ])
        except Exception as internal_e:
            log.error("Error while creating group: {}", internal_e)


def _generate_consumer_id(node_id: Optional[str]) -> str:
    h = hashlib.sha1()
    h.update(str(node_id or socket.getfqdn()).encode("utf8"))
    hostname_hash = h.hexdigest()
    h = hashlib.sha1()
    h.update(__file__.encode("utf8"))
    installation_path_hash = h.hexdigest()
    pidx = process_index.get(0)
    return f"{hostname_hash}:{installation_path_hash}:{pidx}"



import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional, override

from ai.backend.common.bgtask.types import BgtaskStatus
from ai.backend.common.events.types import AbstractEvent, EventDomain
from ai.backend.common.events.user_event.user_bgtask_event import (
    UserBgtaskCancelledEvent,
    UserBgtaskDoneEvent,
    UserBgtaskFailedEvent,
    UserBgtaskUpdatedEvent,
)
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.exception import UnreachableError
from ai.backend.common.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class BaseBgtaskEvent(AbstractEvent, ABC):
    task_id: uuid.UUID

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.BGTASK

    @override
    def domain_id(self) -> Optional[str]:
        return str(self.task_id)

    @abstractmethod
    def status(self) -> BgtaskStatus:
        raise NotImplementedError


@dataclass
class BgtaskUpdatedEvent(BaseBgtaskEvent):
    current_progress: float
    total_progress: float
    message: Optional[str] = None

    def serialize(self) -> tuple:
        return (
            str(self.task_id),
            self.current_progress,
            self.total_progress,
            self.message,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            uuid.UUID(value[0]),
            value[1],
            value[2],
            value[3],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "bgtask_updated"

    @override
    def status(self) -> BgtaskStatus:
        return BgtaskStatus.STARTED

    @override
    def user_event(self) -> Optional[UserEvent]:
        return UserBgtaskUpdatedEvent(
            task_id=str(self.task_id),
            message=str(self.message),
            current_progress=self.current_progress,
            total_progress=self.total_progress,
        )


@dataclass
class BaseBgtaskDoneEvent(BaseBgtaskEvent):
    """
    Arguments for events that are triggered when the Bgtask is completed.
    """

    message: Optional[str] = None

    @override
    def serialize(self) -> tuple:
        return (
            str(self.task_id),
            self.message,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            uuid.UUID(value[0]),
            value[1],
        )


@dataclass
class BgtaskDoneEvent(BaseBgtaskDoneEvent):
    """
    Event triggered when the Bgtask is successfully completed.
    """

    @classmethod
    @override
    def event_name(cls) -> str:
        return "bgtask_done"

    @override
    def status(self) -> BgtaskStatus:
        return BgtaskStatus.DONE

    @override
    def user_event(self) -> Optional[UserEvent]:
        return UserBgtaskDoneEvent(
            task_id=str(self.task_id),
            message=str(self.message),
        )


@dataclass
class BgtaskAlreadyDoneEvent(BaseBgtaskEvent):
    """
    Event triggered when the Bgtask is already completed.
    An event recreated based on the last status of the Bgtask.
    """

    task_status: BgtaskStatus
    message: Optional[str] = None
    current: str = "0"
    total: str = "0"

    @override
    def serialize(self) -> tuple:
        raise UnreachableError("BgtaskAlreadyDoneEvent should not be serialized.")

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        raise UnreachableError("BgtaskAlreadyDoneEvent should not be deserialized.")

    @classmethod
    @override
    def event_name(cls) -> str:
        return "bgtask_already_done"

    @override
    def status(self) -> BgtaskStatus:
        return self.task_status

    @override
    def user_event(self) -> Optional[UserEvent]:
        match self.task_status:
            case BgtaskStatus.DONE:
                return UserBgtaskDoneEvent(
                    task_id=str(self.task_id),
                    message=str(self.message),
                )
            case BgtaskStatus.CANCELLED:
                return UserBgtaskCancelledEvent(
                    task_id=str(self.task_id),
                    message=str(self.message),
                )
            case BgtaskStatus.FAILED:
                return UserBgtaskFailedEvent(
                    task_id=str(self.task_id),
                    message=str(self.message),
                )
            case BgtaskStatus.PARTIAL_SUCCESS:
                return UserBgtaskDoneEvent(
                    task_id=str(self.task_id),
                    message=str(self.message),
                )
            case _:
                log.exception("unknown task status {}", self.task_status)
                raise UnreachableError(f"Unknown task status {self.task_status}")


@dataclass
class BgtaskCancelledEvent(BaseBgtaskDoneEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "bgtask_cancelled"

    @override
    def status(self) -> BgtaskStatus:
        return BgtaskStatus.CANCELLED

    @override
    def user_event(self) -> Optional[UserEvent]:
        return UserBgtaskCancelledEvent(
            task_id=str(self.task_id),
            message=str(self.message),
        )


@dataclass
class BgtaskFailedEvent(BaseBgtaskDoneEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "bgtask_failed"

    @override
    def status(self) -> BgtaskStatus:
        return BgtaskStatus.FAILED

    @override
    def user_event(self) -> Optional[UserEvent]:
        return UserBgtaskFailedEvent(
            task_id=str(self.task_id),
            message=str(self.message),
        )


@dataclass
class BgtaskPartialSuccessEvent(BaseBgtaskDoneEvent):
    errors: list[str] = field(default_factory=list)

    @override
    def serialize(self) -> tuple:
        return (
            str(self.task_id),
            self.message,
            self.errors,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            uuid.UUID(value[0]),
            value[1],
            value[2],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "bgtask_partial_success"

    @override
    def status(self) -> BgtaskStatus:
        # TODO: When client side is ready, we can change this to `TaskStatus.PARTIAL_SUCCESS`
        return BgtaskStatus.DONE

    @override
    def user_event(self) -> Optional[UserEvent]:
        # TODO: When client side is ready, we can change this to `UserBgtaskPartialSuccessEvent`
        return UserBgtaskDoneEvent(
            task_id=str(self.task_id),
            message=str(self.message),
        )



from typing import Optional

from pydantic import Field

from ...api_handlers import BaseResponseModel
from .field import VFolderMetaField, VolumeMetaField


class GetVolumeResponse(BaseResponseModel):
    item: VolumeMetaField = Field(
        description="Provides metadata for a specific volume, used to manage and track storage operations."
    )


class GetVolumesResponse(BaseResponseModel):
    items: list[VolumeMetaField] = Field(
        description="Retrieves metadata for all available volumes, allowing for bulk management and monitoring."
    )


class QuotaScopeResponse(BaseResponseModel):
    used_bytes: Optional[int] = Field(
        default=0,
        description="Indicates the current usage within a quota scope, used for enforcing storage limits.",
    )
    limit_bytes: Optional[int] = Field(
        default=0,
        description="Defines the maximum allowed storage capacity within a quota scope, ensuring controlled resource allocation.",
    )


class VFolderMetadataResponse(BaseResponseModel):
    item: VFolderMetaField = Field(
        description="Provides metadata for a vfolder, used for storage tracking and access management."
    )



import enum
import uuid
from dataclasses import dataclass, field
from typing import Any, Mapping, Optional, Self, override

from ai.backend.common.events.types import AbstractEvent, EventDomain
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.types import ContainerId, KernelId, SessionId


class KernelLifecycleEventReason(enum.StrEnum):
    AGENT_TERMINATION = "agent-termination"
    ALREADY_TERMINATED = "already-terminated"
    ANOMALY_DETECTED = "anomaly-detected"
    EXEC_TIMEOUT = "exec-timeout"
    FAILED_TO_CREATE = "failed-to-create"
    FAILED_TO_START = "failed-to-start"
    FORCE_TERMINATED = "force-terminated"
    BOOTSTRAP_TIMEOUT = "bootstrap-timeout"
    HANG_TIMEOUT = "hang-timeout"
    IDLE_TIMEOUT = "idle-timeout"
    IDLE_SESSION_LIFETIME = "idle-session-lifetime"
    IDLE_UTILIZATION = "idle-utilization"
    KILLED_BY_EVENT = "killed-by-event"
    SERVICE_SCALED_DOWN = "service-scaled-down"
    NEW_CONTAINER_STARTED = "new-container-started"
    PENDING_TIMEOUT = "pending-timeout"
    RESTARTING = "restarting"
    RESTART_TIMEOUT = "restart-timeout"
    RESUMING_AGENT_OPERATION = "resuming-agent-operation"
    SELF_TERMINATED = "self-terminated"
    TASK_FAILED = "task-failed"
    TASK_TIMEOUT = "task-timeout"
    TASK_CANCELLED = "task-cancelled"
    TASK_FINISHED = "task-finished"
    TERMINATED_UNKNOWN_CONTAINER = "terminated-unknown-container"
    UNKNOWN = "unknown"
    USER_REQUESTED = "user-requested"
    NOT_FOUND_IN_MANAGER = "not-found-in-manager"
    CONTAINER_NOT_FOUND = "container-not-found"

    @classmethod
    def from_value(cls, value: Optional[str]) -> Optional[Self]:
        if value is None:
            return None
        try:
            return cls(value)
        except ValueError:
            pass
        return None


@dataclass
class BaseKernelEvent(AbstractEvent):
    kernel_id: KernelId

    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.KERNEL

    @override
    def domain_id(self) -> Optional[str]:
        return str(self.kernel_id)


@dataclass
class KernelLifecycleEvent(BaseKernelEvent):
    session_id: SessionId
    reason: str = ""

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class KernelCreationEvent(KernelLifecycleEvent):
    creation_info: Mapping[str, Any] = field(default_factory=dict)

    @override
    def serialize(self) -> tuple:
        return (
            str(self.kernel_id),
            str(self.session_id),
            self.reason,
            self.creation_info,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            kernel_id=KernelId(uuid.UUID(value[0])),
            session_id=SessionId(uuid.UUID(value[1])),
            reason=value[2],
            creation_info=value[3],
        )

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class KernelPreparingEvent(KernelCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_preparing"


@dataclass
class KernelPullingEvent(KernelCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_pulling"


@dataclass
class KernelCreatingEvent(KernelCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_creating"


class KernelStartedEvent(KernelCreationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_started"


class KernelCancelledEvent(KernelLifecycleEvent):
    @override
    def serialize(self) -> tuple:
        return (
            str(self.kernel_id),
            str(self.session_id),
            self.reason,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            kernel_id=KernelId(uuid.UUID(value[0])),
            session_id=SessionId(uuid.UUID(value[1])),
            reason=value[2],
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_cancelled"


@dataclass
class KernelTerminationEvent(BaseKernelEvent):
    session_id: SessionId
    reason: KernelLifecycleEventReason = KernelLifecycleEventReason.UNKNOWN
    exit_code: int = -1

    @override
    def serialize(self) -> tuple:
        return (
            str(self.kernel_id),
            str(self.session_id),
            self.reason,
            self.exit_code,
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            KernelId(uuid.UUID(value[0])),
            session_id=SessionId(uuid.UUID(value[1])),
            reason=value[2],
            exit_code=value[3],
        )

    @override
    def domain_id(self) -> Optional[str]:
        return None

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


class KernelTerminatingEvent(KernelTerminationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_terminating"


class KernelTerminatedEvent(KernelTerminationEvent):
    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_terminated"


@dataclass
class DoSyncKernelLogsEvent(BaseKernelEvent):
    container_id: str

    def serialize(self) -> tuple:
        return (
            str(self.kernel_id),
            self.container_id,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            KernelId(uuid.UUID(value[0])),
            value[1],
        )

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None

    @classmethod
    def event_name(cls) -> str:
        return "do_sync_kernel_logs"


@dataclass
class KernelHeartbeatEvent(BaseKernelEvent):
    container_id: ContainerId

    @override
    def serialize(self) -> tuple:
        return (
            str(self.kernel_id),
            str(self.container_id),
        )

    @classmethod
    @override
    def deserialize(cls, value: tuple):
        return cls(
            kernel_id=KernelId(uuid.UUID(value[0])),
            container_id=ContainerId(str(value[1])),
        )

    @classmethod
    @override
    def event_name(cls) -> str:
        return "kernel_heartbeat"

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Callable, Final

from aiomonitor.task import preserve_termination_log

from ai.backend.logging import BraceStyleAdapter

if TYPE_CHECKING:
    from .events.dispatcher import AbstractEvent, EventProducer
    from .lock import AbstractDistributedLock


log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GlobalTimer:
    """
    Executes the given async function only once in the given interval,
    uniquely among multiple manager instances across multiple nodes.
    """

    _event_producer: Final[EventProducer]

    def __init__(
        self,
        dist_lock: AbstractDistributedLock,
        event_producer: EventProducer,
        event_factory: Callable[[], AbstractEvent],
        interval: float = 10.0,
        initial_delay: float = 0.0,
        *,
        task_name: str | None = None,
    ) -> None:
        self._dist_lock = dist_lock
        self._event_producer = event_producer
        self._event_factory = event_factory
        self._stopped = False
        self.interval = interval
        self.initial_delay = initial_delay
        self.task_name = task_name

    @preserve_termination_log
    async def generate_tick(self) -> None:
        try:
            await asyncio.sleep(self.initial_delay)
            if self._stopped:
                return
            while True:
                try:
                    async with self._dist_lock:
                        if self._stopped:
                            return
                        await self._event_producer.produce_event(self._event_factory())
                        if self._stopped:
                            return
                        await asyncio.sleep(self.interval)
                except asyncio.TimeoutError:  # timeout raised from etcd lock
                    if self._stopped:
                        return
                    log.warning("timeout raised while trying to acquire lock. retrying...")
        except asyncio.CancelledError:
            pass

    async def join(self) -> None:
        self._tick_task = asyncio.create_task(self.generate_tick())
        if self.task_name is not None:
            self._tick_task.set_name(self.task_name)

    async def leave(self) -> None:
        self._stopped = True
        await asyncio.sleep(0)
        if not self._tick_task.done():
            try:
                self._tick_task.cancel()
                await self._tick_task
            except asyncio.CancelledError:
                pass



from dataclasses import dataclass
from typing import Optional

from ai.backend.common.typed_validators import HostPortPair


@dataclass
class EtcdConfigData:
    namespace: str
    addr: HostPortPair
    user: Optional[str]
    password: Optional[str]



import asyncio
import logging
import uuid
from typing import AsyncIterator, Optional, Protocol

from ai.backend.common.bgtask.bgtask import BgTaskInfo
from ai.backend.logging.utils import BraceStyleAdapter

from ...bgtask import BgtaskAlreadyDoneEvent
from ...dispatcher import AbstractEvent
from ..hub import EventPropagator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class BgtaskLastDoneEventFetcher(Protocol):
    async def fetch_bgtask_info(
        self,
        task_id: uuid.UUID,
    ) -> BgTaskInfo:
        """
        Fetch the background task information for a given task ID.
        This method is used to retrieve the last status of a background task.
        """
        ...


class BgtaskPropagator(EventPropagator):
    """
    Background task propagator that uses an asyncio queue to propagate events.
    This propagator is used to handle events related to background tasks.
    """

    _id: uuid.UUID
    _last_done_event_fetcher: BgtaskLastDoneEventFetcher
    _queue: asyncio.Queue[Optional[AbstractEvent]]
    _closed: bool = False

    def __init__(self, last_done_event_fetcher: BgtaskLastDoneEventFetcher) -> None:
        self._id = uuid.uuid4()
        self._last_done_event_fetcher = last_done_event_fetcher
        self._queue = asyncio.Queue()
        self._closed = False

    def id(self) -> uuid.UUID:
        """
        Get the unique identifier for the propagator.
        """
        return self._id

    async def receive(self, task_id: uuid.UUID) -> AsyncIterator[AbstractEvent]:
        """
        Receive background task events from the queue.
        First, it fetches the last finished event for a given task ID.
        If the last event is not None, it yields that event.
        Then, it enters a loop to receive events from the queue until closed.
        """
        bgtask_info = await self._last_done_event_fetcher.fetch_bgtask_info(task_id)
        if bgtask_info.status.finished():
            log.debug(
                "Yielding already finished event: {}",
                bgtask_info,
            )
            yield BgtaskAlreadyDoneEvent(
                task_id=task_id,
                message=bgtask_info.msg,
                task_status=bgtask_info.status,
                current=bgtask_info.current,
                total=bgtask_info.total,
            )
            return
        while not self._closed:
            event = await self._queue.get()
            try:
                if event is None:
                    break
                yield event
            except Exception as e:
                log.error("Error propagating event: {}", e)

    async def propagate_event(self, event: AbstractEvent) -> None:
        await self._queue.put(event)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self._queue.put(None)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.events.types import AbstractEvent, EventDomain
from ai.backend.common.events.user_event.user_event import UserEvent
from ai.backend.common.types import QuotaScopeID, VolumeMountableNodeType


class BaseVolumeEvent(AbstractEvent):
    @classmethod
    @override
    def event_domain(cls) -> EventDomain:
        return EventDomain.VOLUME

    @override
    def domain_id(self) -> Optional[str]:
        return None

    @override
    def user_event(self) -> Optional[UserEvent]:
        return None


@dataclass
class DoVolumeMountEvent(BaseVolumeEvent):
    # Let storage proxies and agents find the real path of volume
    # with their mount_path or mount_prefix.
    dir_name: str
    volume_backend_name: str
    quota_scope_id: QuotaScopeID

    fs_location: str
    fs_type: str = "nfs"
    cmd_options: Optional[str] = None
    scaling_group: Optional[str] = None

    # if `edit_fstab` is False, `fstab_path` is ignored
    # if `edit_fstab` is True, `fstab_path` or "/etc/fstab" is used to edit fstab
    edit_fstab: bool = False
    fstab_path: str = "/etc/fstab"

    def serialize(self) -> tuple:
        return (
            self.dir_name,
            self.volume_backend_name,
            str(self.quota_scope_id),
            self.fs_location,
            self.fs_type,
            self.cmd_options,
            self.scaling_group,
            self.edit_fstab,
            self.fstab_path,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            dir_name=value[0],
            volume_backend_name=value[1],
            quota_scope_id=QuotaScopeID.parse(value[2]),
            fs_location=value[3],
            fs_type=value[4],
            cmd_options=value[5],
            scaling_group=value[6],
            edit_fstab=value[7],
            fstab_path=value[8],
        )

    @classmethod
    def event_name(cls) -> str:
        return "do_volume_mount"


@dataclass
class DoVolumeUnmountEvent(BaseVolumeEvent):
    # Let storage proxies and agents find the real path of volume
    # with their mount_path or mount_prefix.
    dir_name: str
    volume_backend_name: str
    quota_scope_id: QuotaScopeID
    scaling_group: Optional[str] = None

    # if `edit_fstab` is False, `fstab_path` is ignored
    # if `edit_fstab` is True, `fstab_path` or "/etc/fstab" is used to edit fstab
    edit_fstab: bool = False
    fstab_path: Optional[str] = None

    def serialize(self) -> tuple:
        return (
            self.dir_name,
            self.volume_backend_name,
            str(self.quota_scope_id),
            self.scaling_group,
            self.edit_fstab,
            self.fstab_path,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            dir_name=value[0],
            volume_backend_name=value[1],
            quota_scope_id=QuotaScopeID.parse(value[2]),
            scaling_group=value[3],
            edit_fstab=value[4],
            fstab_path=value[5],
        )

    @classmethod
    def event_name(cls) -> str:
        return "do_volume_unmount"


@dataclass
class BaseAgentVolumeMountEvent(BaseVolumeEvent):
    node_id: str
    node_type: VolumeMountableNodeType
    mount_path: str
    quota_scope_id: QuotaScopeID
    err_msg: Optional[str] = None

    def serialize(self) -> tuple:
        return (
            self.node_id,
            str(self.node_type),
            self.mount_path,
            str(self.quota_scope_id),
            self.err_msg,
        )

    @classmethod
    def deserialize(cls, value: tuple):
        return cls(
            value[0],
            VolumeMountableNodeType(value[1]),
            value[2],
            QuotaScopeID.parse(value[3]),
            value[4],
        )


class VolumeMounted(BaseAgentVolumeMountEvent):
    @classmethod
    def event_name(cls) -> str:
        return "volume_mounted"


class VolumeUnmounted(BaseAgentVolumeMountEvent):
    @classmethod
    def event_name(cls) -> str:
        return "volume_unmounted"



import enum
from abc import ABC, abstractmethod
from typing import Optional, Self

from .user_event.user_event import UserEvent

__all__ = (
    "EventDomain",
    "AbstractEvent",
)


class EventDomain(enum.StrEnum):
    BGTASK = "bgtask"
    IMAGE = "image"
    KERNEL = "kernel"
    MODEL_SERVING = "model_serving"
    MODEL_ROUTE = "model_route"
    SCHEDULE = "schedule"
    IDLE_CHECK = "idle_check"
    SESSION = "session"
    AGENT = "agent"
    VFOLDER = "vfolder"
    VOLUME = "volume"
    LOG = "log"


class AbstractEvent(ABC):
    @abstractmethod
    def serialize(self) -> tuple[bytes, ...]:
        """
        Return a msgpack-serializable tuple.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def deserialize(cls, value: tuple[bytes, ...]) -> Self:
        """
        Construct the event args from a tuple deserialized from msgpack.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def event_domain(self) -> EventDomain:
        """
        Return the event domain.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def event_name(cls) -> str:
        """
        Return the event name.
        """
        raise NotImplementedError

    @abstractmethod
    def domain_id(self) -> Optional[str]:
        """
        Return the domain ID.
        It's used to identify the event domain in the event hub.
        """
        raise NotImplementedError

    @abstractmethod
    def user_event(self) -> Optional[UserEvent]:
        """
        Return the event as a UserEvent.
        If user event is not supported, return None.
        """
        raise NotImplementedError



from __future__ import annotations

import asyncio
import enum
import logging
import secrets
import time
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    Optional,
    Protocol,
    Type,
    TypedDict,
    TypeVar,
    Union,
    cast,
    override,
)

import attrs
from aiomonitor.task import preserve_termination_log
from aiotools.taskgroup import PersistentTaskGroup
from aiotools.taskgroup.types import AsyncExceptionHandler

from ai.backend.common.message_queue.queue import AbstractMessageQueue, MessageId
from ai.backend.logging import BraceStyleAdapter

from .. import msgpack
from ..types import (
    AgentId,
)
from .reporter import AbstractEventReporter, CompleteEventReportArgs, PrepareEventReportArgs
from .types import AbstractEvent

__all__ = (
    "EventCallback",
    "EventDispatcher",
    "EventHandler",
    "EventProducer",
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class _EventHandlerType(enum.StrEnum):
    CONSUMER = "consumer"
    SUBSCRIBER = "subscriber"


TEvent = TypeVar("TEvent", bound="AbstractEvent")
TEventCov = TypeVar("TEventCov", bound="AbstractEvent")
TContext = TypeVar("TContext")

EventCallback = Union[
    Callable[[TContext, AgentId, TEvent], Coroutine[Any, Any, None]],
    Callable[[TContext, AgentId, TEvent], None],
]


class EventHandlerType(enum.Enum):
    CONSUMER = "CONSUMER"
    SUBSCRIBER = "SUBSCRIBER"


@attrs.define(auto_attribs=True, slots=True, frozen=True, eq=False, order=False)
class EventHandler(Generic[TContext, TEvent]):
    event_cls: Type[TEvent]
    name: str
    context: TContext
    callback: EventCallback[TContext, TEvent]
    handler_type: _EventHandlerType
    coalescing_opts: Optional[CoalescingOptions]
    coalescing_state: CoalescingState
    args_matcher: Callable[[tuple], bool] | None
    event_start_reporters: tuple[AbstractEventReporter, ...] = attrs.field(factory=tuple)
    event_complete_reporters: tuple[AbstractEventReporter, ...] = attrs.field(factory=tuple)


class CoalescingOptions(TypedDict):
    max_wait: float
    max_batch_size: int


@attrs.define(auto_attribs=True, slots=True)
class CoalescingState:
    batch_size: int = 0
    last_added: float = 0.0
    last_handle: asyncio.TimerHandle | None = None
    fut_sync: asyncio.Future | None = None

    def proceed(self):
        if self.fut_sync is not None and not self.fut_sync.done():
            self.fut_sync.set_result(None)

    async def rate_control(self, opts: CoalescingOptions | None) -> bool:
        if opts is None:
            return True
        loop = asyncio.get_running_loop()
        if self.fut_sync is None:
            self.fut_sync = loop.create_future()
        assert self.fut_sync is not None
        self.last_added = loop.time()
        self.batch_size += 1
        if self.batch_size >= opts["max_batch_size"]:
            assert self.last_handle is not None
            self.last_handle.cancel()
            self.fut_sync.cancel()
            self.last_handle = None
            self.last_added = 0.0
            self.batch_size = 0
            return True
        # Schedule.
        self.last_handle = loop.call_later(
            opts["max_wait"],
            self.proceed,
        )
        if self.last_added > 0 and loop.time() - self.last_added < opts["max_wait"]:
            # Cancel the previously pending task.
            self.last_handle.cancel()
            self.fut_sync.cancel()
            # Reschedule.
            self.fut_sync = loop.create_future()
            self.last_handle = loop.call_later(
                opts["max_wait"],
                self.proceed,
            )
        try:
            await self.fut_sync
        except asyncio.CancelledError:
            if self.last_handle is not None and not self.last_handle.cancelled():
                self.last_handle.cancel()
            return False
        else:
            self.fut_sync = None
            self.last_handle = None
            self.last_added = 0.0
            self.batch_size = 0
            return True


class EventObserver(Protocol):
    def observe_event_success(self, *, event_type: str, duration: float) -> None: ...

    def observe_event_failure(
        self, *, event_type: str, duration: float, exception: BaseException
    ) -> None: ...


class NopEventObserver:
    def observe_event_success(self, *, event_type: str, duration: float) -> None:
        pass

    def observe_event_failure(
        self, *, event_type: str, duration: float, exception: BaseException
    ) -> None:
        pass


class _ConsumerPostCallback:
    def __init__(
        self,
        msg_id: MessageId,
        msg_queue: AbstractMessageQueue,
        remaining_handler_cnt: int,
    ) -> None:
        self._msg_id = msg_id
        self._msg_queue = msg_queue
        self._remaining_handler_cnt = remaining_handler_cnt
        self._lock = asyncio.Lock()

    async def done(self) -> None:
        # To ensure that all consumer handlers are called.
        # Basically there should be only one consumer handler for one event.
        async with self._lock:
            self._remaining_handler_cnt -= 1
            if self._remaining_handler_cnt > 0:
                return
        # All consumer handlers are called.
        await self._msg_queue.done(self._msg_id)


class PostCallback(Protocol):
    async def done(self) -> None:
        pass


class EventDispatcherGroup(ABC):
    @abstractmethod
    def with_reporters(
        self,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventDispatcherGroup:
        raise NotImplementedError

    @abstractmethod
    def consume(
        self,
        event_cls: Type[TEvent],
        context: TContext,
        callback: EventCallback[TContext, TEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TEvent]:
        raise NotImplementedError

    @abstractmethod
    def subscribe(
        self,
        event_cls: Type[TEvent],
        context: TContext,
        callback: EventCallback[TContext, TEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        override_event_name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TEvent]:
        raise NotImplementedError


class _EventDispatcherWrapper(EventDispatcherGroup):
    _event_dispatcher: EventDispatcher

    _start_reporters: list[AbstractEventReporter]
    _complete_reporters: list[AbstractEventReporter]

    def __init__(
        self,
        event_dispatcher: EventDispatcher,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> None:
        self._event_dispatcher = event_dispatcher
        self._start_reporters = list(start_reporters)
        self._complete_reporters = list(complete_reporters)

    @override
    def with_reporters(
        self,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> _EventDispatcherWrapper:
        return _EventDispatcherWrapper(
            event_dispatcher=self._event_dispatcher,
            start_reporters=self._start_reporters + list(start_reporters),
            complete_reporters=self._complete_reporters + list(complete_reporters),
        )

    @override
    def consume(
        self,
        event_cls: Type[TEvent],
        context: TContext,
        callback: EventCallback[TContext, TEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TEvent]:
        return self._event_dispatcher.consume(
            event_cls,
            context,
            callback,
            coalescing_opts=coalescing_opts,
            name=name,
            args_matcher=args_matcher,
            start_reporters=tuple(self._start_reporters),
            complete_reporters=tuple(self._complete_reporters),
        )

    @override
    def subscribe(
        self,
        event_cls: Type[TEvent],
        context: TContext,
        callback: EventCallback[TContext, TEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        override_event_name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
    ) -> EventHandler[TContext, TEvent]:
        return self._event_dispatcher.subscribe(
            event_cls,
            context,
            callback,
            coalescing_opts=coalescing_opts,
            name=name,
            override_event_name=override_event_name,
            args_matcher=args_matcher,
            start_reporters=tuple(self._start_reporters),
            complete_reporters=tuple(self._complete_reporters),
        )


class EventDispatcher(EventDispatcherGroup):
    """
    We have two types of event handlers: consumer and subscriber.

    Consumers use the distribution pattern. Only one consumer among many manager worker processes
    receives the event.

    Consumer example: database updates upon specific events.

    Subscribers use the broadcast pattern. All subscribers in many manager worker processes
    receive the same event.

    Subscriber example: enqueuing events to the queues for event streaming API handlers
    """

    _consumers: defaultdict[
        str, set[EventHandler[Any, AbstractEvent]]
    ]  # TODO: set only one consumer handler for one event
    _subscribers: defaultdict[str, set[EventHandler[Any, AbstractEvent]]]
    _msg_queue: AbstractMessageQueue

    _consumer_loop_task: Optional[asyncio.Task]
    _subscriber_loop_task: Optional[asyncio.Task]
    _consumer_taskgroup: PersistentTaskGroup
    _subscriber_taskgroup: PersistentTaskGroup

    _log_events: bool
    _metric_observer: EventObserver

    def __init__(
        self,
        message_queue: AbstractMessageQueue,
        log_events: bool = False,
        *,
        consumer_exception_handler: AsyncExceptionHandler | None = None,
        subscriber_exception_handler: AsyncExceptionHandler | None = None,
        event_observer: EventObserver = NopEventObserver(),
    ) -> None:
        self._log_events = log_events
        self._closed = False
        self._consumers = defaultdict(set)
        self._subscribers = defaultdict(set)
        self._msg_queue = message_queue
        self._metric_observer = event_observer
        self._consumer_taskgroup = PersistentTaskGroup(
            name="consumer_taskgroup",
            exception_handler=consumer_exception_handler,
        )
        self._subscriber_taskgroup = PersistentTaskGroup(
            name="subscriber_taskgroup",
            exception_handler=subscriber_exception_handler,
        )
        self._consumer_loop_task = None
        self._subscriber_loop_task = None

    async def start(self) -> None:
        if self._closed:
            return
        self._consumer_loop_task = asyncio.create_task(self._consume_loop())
        self._subscriber_loop_task = asyncio.create_task(self._subscribe_loop())

    async def close(self) -> None:
        self._closed = True
        try:
            cancelled_tasks = []
            await self._consumer_taskgroup.shutdown()
            await self._subscriber_taskgroup.shutdown()

            def cancel_task(task: Optional[asyncio.Task]) -> None:
                if task is not None and not task.done():
                    task.cancel()
                    cancelled_tasks.append(task)

            cancel_task(self._consumer_loop_task)
            cancel_task(self._subscriber_loop_task)
            await asyncio.gather(*cancelled_tasks, return_exceptions=True)
        except Exception:
            log.exception("unexpected error while closing event dispatcher")

    @override
    def with_reporters(
        self,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventDispatcherGroup:
        return _EventDispatcherWrapper(
            event_dispatcher=self,
            start_reporters=list(start_reporters),
            complete_reporters=list(complete_reporters),
        )

    @override
    def consume(
        self,
        event_cls: Type[TEvent],
        context: TContext,
        callback: EventCallback[TContext, TEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventHandler[TContext, TEvent]:
        """
        Register a callback as a consumer. When multiple callback registers as a consumer
        on a single event, only one callable among those will be called.

        args_matcher:
          Optional. A callable which accepts event argument and supplies a bool as a return value.
          When specified, EventDispatcher will only execute callback when this lambda returns True.
        """

        if name is None:
            name = f"evh-{secrets.token_urlsafe(16)}"
        handler = EventHandler(
            event_cls,
            name,
            context,
            callback,
            _EventHandlerType.CONSUMER,
            coalescing_opts,
            CoalescingState(),
            args_matcher,
            event_start_reporters=tuple(start_reporters),
            event_complete_reporters=tuple(complete_reporters),
        )
        self._consumers[event_cls.event_name()].add(cast(EventHandler[Any, AbstractEvent], handler))
        return handler

    def unconsume(
        self,
        handler: EventHandler[TContext, TEvent],
    ) -> None:
        self._consumers[handler.event_cls.event_name()].discard(
            cast(EventHandler[Any, AbstractEvent], handler)
        )

    @override
    def subscribe(
        self,
        event_cls: Type[TEvent],
        context: TContext,
        callback: EventCallback[TContext, TEvent],
        coalescing_opts: Optional[CoalescingOptions] = None,
        *,
        name: Optional[str] = None,
        override_event_name: Optional[str] = None,
        args_matcher: Optional[Callable[[tuple], bool]] = None,
        start_reporters: Sequence[AbstractEventReporter] = tuple(),
        complete_reporters: Sequence[AbstractEventReporter] = tuple(),
    ) -> EventHandler[TContext, TEvent]:
        """
        Subscribes to given event. All handlers will be called when certain event pops up.

        args_matcher:
          Optional. A callable which accepts event argument and supplies a bool as a return value.
          When specified, EventDispatcher will only execute callback when this lambda returns True.
        """

        if name is None:
            name = f"evh-{secrets.token_urlsafe(16)}"
        handler = EventHandler(
            event_cls,
            name,
            context,
            callback,
            _EventHandlerType.SUBSCRIBER,
            coalescing_opts,
            CoalescingState(),
            args_matcher,
            event_start_reporters=tuple(start_reporters),
            event_complete_reporters=tuple(complete_reporters),
        )
        override_event_name = override_event_name or event_cls.event_name()
        self._subscribers[override_event_name].add(cast(EventHandler[Any, AbstractEvent], handler))
        return handler

    def unsubscribe(
        self,
        handler: EventHandler[TContext, TEvent],
        *,
        override_event_name: Optional[str] = None,
    ) -> None:
        override_event_name = override_event_name or handler.event_cls.event_name()
        self._subscribers[override_event_name].discard(
            cast(EventHandler[Any, AbstractEvent], handler)
        )

    async def _handle(
        self,
        evh: EventHandler,
        source: AgentId,
        args: tuple,
        post_callbacks: Sequence[PostCallback] = tuple(),
    ) -> None:
        if evh.args_matcher and not evh.args_matcher(args):
            return
        coalescing_opts = evh.coalescing_opts
        coalescing_state = evh.coalescing_state
        cb = evh.callback
        evh_type = evh.handler_type
        event_cls = evh.event_cls
        if self._closed:
            return
        event_type = event_cls.event_name()
        event = event_cls.deserialize(args)
        start = time.perf_counter()
        for start_reporter in evh.event_start_reporters:
            await start_reporter.prepare_event_report(event, PrepareEventReportArgs())
        try:
            if await coalescing_state.rate_control(coalescing_opts):
                if self._closed:
                    return
                if self._log_events:
                    log.debug("DISPATCH_{}(evh:{})", evh_type.name, evh.name)
                if asyncio.iscoroutinefunction(cb):
                    # mypy cannot catch the meaning of asyncio.iscoroutinefunction().
                    await cb(evh.context, source, event)  # type: ignore
                else:
                    cb(evh.context, source, event)  # type: ignore
                for post_callback in post_callbacks:
                    await post_callback.done()
                self._metric_observer.observe_event_success(
                    event_type=event_type,
                    duration=time.perf_counter() - start,
                )
        except Exception as e:
            self._metric_observer.observe_event_failure(
                event_type=event_type,
                duration=time.perf_counter() - start,
                exception=e,
            )
            log.exception(f"EventDispatcher.{evh_type}(): unexpected-error, {repr(e)}")
            raise
        except BaseException as e:
            self._metric_observer.observe_event_failure(
                event_type=event_type,
                duration=time.perf_counter() - start,
                exception=e,
            )
            raise
        duration = time.perf_counter() - start
        for complete in evh.event_complete_reporters:
            await complete.complete_event_report(event, CompleteEventReportArgs(duration))

    async def dispatch_consumers(
        self,
        event_name: str,
        source: AgentId,
        args: tuple,
        post_callbacks: Sequence[PostCallback] = tuple(),
    ) -> None:
        if self._log_events:
            log.debug("DISPATCH_CONSUMERS(ev:{}, ag:{})", event_name, source)
        for consumer in self._consumers[event_name].copy():
            self._consumer_taskgroup.create_task(
                self._handle(consumer, source, args, post_callbacks),
            )
            await asyncio.sleep(0)

    async def dispatch_subscribers(
        self,
        event_name: str,
        source: AgentId,
        args: tuple,
    ) -> None:
        if self._log_events:
            log.debug("DISPATCH_SUBSCRIBERS(ev:{}, ag:{})", event_name, source)
        for subscriber in self._subscribers[event_name].copy():
            self._subscriber_taskgroup.create_task(
                self._handle(subscriber, source, args),
            )
            await asyncio.sleep(0)

    @preserve_termination_log
    async def _consume_loop(self) -> None:
        async for msg in self._msg_queue.consume_queue():  # type: ignore
            if self._closed:
                return
            decoded_event_name = msg.payload[b"name"].decode()
            post_callback = _ConsumerPostCallback(
                msg.msg_id,
                self._msg_queue,
                len(self._consumers[decoded_event_name]),
            )

            await self.dispatch_consumers(
                decoded_event_name,
                AgentId(msg.payload[b"source"].decode()),
                msgpack.unpackb(msg.payload[b"args"]),
                [post_callback],
            )

    @preserve_termination_log
    async def _subscribe_loop(self) -> None:
        async for msg in self._msg_queue.subscribe_queue():  # type: ignore
            if self._closed:
                return
            decoded_event_name = msg.payload[b"name"].decode()
            await self.dispatch_subscribers(
                decoded_event_name,
                AgentId(msg.payload[b"source"].decode()),
                msgpack.unpackb(msg.payload[b"args"]),
            )


class EventProducer:
    _closed: bool
    _msg_queue: AbstractMessageQueue
    _log_events: bool

    def __init__(
        self,
        msg_queue: AbstractMessageQueue,
        *,
        source: AgentId,
        log_events: bool = False,
    ) -> None:
        self._closed = False
        self._msg_queue = msg_queue
        self._source_bytes = source.encode()
        self._log_events = log_events

    async def close(self) -> None:
        self._closed = True
        await self._msg_queue.close()

    async def produce_event(
        self,
        event: AbstractEvent,
        source_override: Optional[AgentId] = None,
    ) -> None:
        if self._closed:
            return
        source_bytes = self._source_bytes
        if source_override is not None:
            source_bytes = source_override.encode()

        raw_event = {
            b"name": event.event_name().encode(),
            b"source": source_bytes,
            b"args": msgpack.packb(event.serialize()),
        }
        await self._msg_queue.send(raw_event)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import AsyncGenerator

type MessageId = bytes
_DEFAULT_RETRY_FIELD = b"_retry_count"
_DEFAULT_MAX_RETRIES = 3


@dataclass
class MQMessage:
    msg_id: MessageId
    payload: dict[bytes, bytes]

    def retry(self) -> bool:
        """
        Retry the message.
        If the message has been retried more than the maximum number of retries,
        the message will be discarded.
        The retry count is stored in the message payload.
        """
        if self._retry_count() > _DEFAULT_MAX_RETRIES:
            return False
        self.payload[_DEFAULT_RETRY_FIELD] = str(self._retry_count() + 1).encode("utf-8")
        return True

    def _retry_count(self) -> int:
        """
        Get the retry count of the message.
        The retry count is the number of times the message has been re-delivered.
        """
        return int(self.payload.get(_DEFAULT_RETRY_FIELD, b"0"))


class AbstractMessageQueue(ABC):
    @abstractmethod
    async def send(
        self,
        payload: dict[bytes, bytes],
    ) -> None:
        """
        Send a message to the queue.

        If the queue is full, the oldest message will be removed.
        The new message will be added to the end of the queue.
        The message will be delivered to one consumer.
        """

    @abstractmethod
    async def consume_queue(
        self,
    ) -> AsyncGenerator[MQMessage, None]:
        """
        Consume messages from the queue.
        This method will block until a message is available.

        This is a normal queue, so the message will be delivered to one consumer.
        Messages are consumed only once by one consumer.
        """

    @abstractmethod
    async def subscribe_queue(
        self,
    ) -> AsyncGenerator[MQMessage, None]:
        """
        Subscribe to messages from the queue.
        This method will block until a message is available.

        This is a broadcast queue, so the message will be delivered to all subscribers.
        The subscriber should call `done` method to acknowledge the message when it is processed.
        """

    @abstractmethod
    async def done(
        self,
        msg_id: MessageId,
    ) -> None:
        """
        Acknowledge the message.

        This method should be called after the message is processed.
        If the consumer does not call `done`, the message will be re-delivered after the
        some timeout period.
        """

    @abstractmethod
    async def close(self) -> None:
        """
        Close the message queue.

        This method should be called when the message queue is no longer needed.
        It will close all connections and stop all background tasks.
        """
